from collections import Counter
from html.parser import HTMLParser
from random import sample
import re
from typing import Any, Dict, Optional, Tuple, Union, List
import warnings

from .get import get
from .utils import match, recover_html_and_attrs


class Soup(HTMLParser):
    """\
    HTML Soup Parser

    Attributes:

    - html: content to parse
    - tag: element to match
    - attrs: element attributes to match
    - text: inner data

    Methods:

    - find: matching content by element tag (and attributes)
    - strip: brackets, tags, and attributes from inner data
    - get: alternate initializer

    Deprecations:

    - remove_tags: (as of 1.0) use strip

    Examples:

    ```
    from gazpacho import Soup

    html = "<div><p class='a'>1</p><p class='a'>2</p><p class='b'>3</p></div>"
    url = "https://www.gazpacho.xyz"

    soup = Soup(html)
    soup = Soup.get(url)
    ```
    """

    def __init__(self, html: Optional[str] = None) -> None:
        """\
        Arguments:

        - html: content to parse
        """
        super().__init__()
        self.html = "" if not html else html
        self.tag: Optional[str] = None
        self.attrs: Optional[Dict[Any, Any]] = None
        self.text: Optional[str] = None

    def __dir__(self):
        return ["attrs", "find", "get", "html", "strip", "tag", "text"]

    def __repr__(self) -> str:
        return self.html

    @classmethod
    def get(
        cls,
        url: str,
        params: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> "Soup":
        """\
        Intialize with gazpacho.get
        """
        html = get(url, params, headers)
        if not isinstance(html, str):
            raise Exception(f"Unable to retrieve contents from {url}")
        return cls(html)

    @property
    def _active(self) -> bool:
        return sum(self.counter.values()) > 0

    @staticmethod
    def _void(tag: str) -> bool:
        return tag in [
            "area",
            "base",
            "br",
            "col",
            "embed",
            "hr",
            "img",
            "input",
            "keygen",
            "link",
            "meta",
            "param",
            "source",
            "track",
            "wbr",
        ]

    def _handle_start(self, tag: str, attrs: List[Tuple[str, Optional[str]]]) -> None:
        html, attrs_dict = recover_html_and_attrs(tag, attrs)
        query_attrs = {} if not self.attrs else self.attrs
        matching = match(query_attrs, attrs_dict, partial=self.partial)

        if (tag == self.tag) and (matching) and (not self._active):
            self.groups.append(Soup())
            self.groups[-1].tag = tag
            self.groups[-1].attrs = attrs_dict
            self.groups[-1].html += html
            self.counter[tag] += 1
            return

        if self._active:
            self.groups[-1].html += html
            self.counter[tag] += 1

    def handle_starttag(self, tag: str, attrs: List[Tuple[str, Optional[str]]]) -> None:
        self._handle_start(tag, attrs)
        if self._active:
            if self._void(tag):
                self.counter[tag] -= 1

    def handle_startendtag(
        self, tag: str, attrs: List[Tuple[str, Optional[str]]]
    ) -> None:
        self._handle_start(tag, attrs)
        if self._active:
            self.counter[tag] -= 1

    def handle_data(self, data: str) -> None:
        if self._active:
            if self.groups[-1].text is None:
                self.groups[-1].text = data.strip()
            self.groups[-1].html += data

    def handle_endtag(self, tag: str) -> None:
        if self._active:
            self.groups[-1].html += f"</{tag}>"
            self.counter[tag] -= 1

    def strip(self, whitespace: bool = True) -> str:
        """\
        Strip brackets, tags, and attributes from inner text

        Arguments:

        - whitespace: remove extra whitespace characters

        Example:

        ```
        html = "<span>AB<b>C</b>D</span>"
        soup = Soup(html)
        soup.find("span").text
        # AB
        soup.strip()
        # ABCD
        ```
        """
        text = re.sub("<[^>]+>", "", self.html)
        if whitespace:
            text = " ".join(text.split())
        return text

    def remove_tags(self, strip: bool = True) -> str:
        """\
        Now: .strip()
        """
        message = "Marked for removal; use .strip()"
        warnings.warn(message, category=FutureWarning, stacklevel=2)
        return self.strip(whitespace=strip)

    def find(
        self,
        tag: str,
        attrs: Optional[Dict[str, str]] = None,
        *,
        partial: bool = True,
        mode: str = "automatic",
        strict: Optional[bool] = None,
    ) -> Optional[Union[List["Soup"], "Soup"]]:
        """\
        Return matching HTML elements

        Arguments:

        - tag: target element tag
        - attrs: target element attributes
        - partial: match on attributes
        - mode: override return behavior {'auto/automatic', 'all/list', 'first'}

        Deprecations:

        - strict: (as of 1.0) use partial=

        Examples:

        ```
        soup.find('p', {'class': 'a'})
        # [<p class="a">1</p>, <p class="a">2</p>]

        soup.find('p', {'class': 'a'}, mode='first')
        # <p class="a">1</p>

        result = soup.find('p', {'class': 'b'}, mode='auto')
        print(result)
        # <p class="b">3</p>

        print(result.text)
        # 3
        ```
        """
        self.counter: Counter = Counter()
        self.groups: List = []
        self.tag = tag
        self.attrs = attrs
        self.partial = partial

        if strict is not None:
            message = "Marked for removal; use partial="
            warnings.warn(message, category=FutureWarning, stacklevel=2)
            partial = not strict

        self.feed(self.html)

        automatic = ["auto", "automatic"]
        all = ["all", "list"]
        first = ["first"]
        last = ["last"]  # undocumented
        random = ["random"]  # undocumented

        if not self.groups:
            if mode in all:
                return []
            else:
                return None
        elif mode in automatic:
            if len(self.groups) == 1:
                return self.groups[0]
            else:
                return self.groups
        elif mode in all:
            return self.groups
        elif mode in first:
            return self.groups[0]
        elif mode in last:
            return self.groups[-1]
        elif mode in random:
            return sample(self.groups, k=1)[0]
        else:
            raise ValueError(mode)



from typing import Any, Dict, List, Optional, Tuple, Union


def match(a: Dict[Any, Any], b: Dict[Any, Any], *, partial: bool = False) -> bool:
    """Match two dictionaries

    Arguments:

    - a: query dict
    - b: dict to match
    - partial: allow partial match

    Examples:

    ```
    a = {'foo': 'bar'}
    b = {'foo': 'bar baz'}
    match(a, b, partial=True)
    # True

    a = {'foo': 'bar'}
    b = {'foo': 'bar baz'}
    match(a, b)
    # False

    a = {}
    b = {'foo': 'bar'}
    match(a, b)
    # True

    a = {}
    b = {}
    match(a, b)
    # True
    ```
    """
    if (not a) or (not a and not b):
        return True
    if a and (not b):
        return False
    for key, lhs in a.items():
        rhs = b.get(key)
        if not rhs:
            return False
        if not partial:
            if lhs == rhs:
                continue
            else:
                return False
        if lhs in rhs:
            continue
        else:
            return False
    return True


def recover_html_and_attrs(
    tag: str, attrs: List[Tuple[str, Optional[str]]], startendtag: bool = False
) -> Tuple[str, Dict[Any, Any]]:
    """\
    Recover html and attrs from HTMLParser feed

    Arguments:

    - tag: element tag
    - attrs: element attributes
    - startendtag: if startend tag

    Example:

    ```
    recover_html_and_attrs('img', [('src', 'example.png')])
    # ("<img src='example.png'>", {'src': 'example.png'})
    ```
    """
    if attrs:
        attrs_dict = dict(attrs)
        attrs_list = [f'{key}="{value}"' for key, value in attrs_dict.items()]
        attrs_str = f' {" ".join(attrs_list)}'
    else:
        attrs_dict = {}
        attrs_str = ""
    if startendtag:
        html = f"<{tag}{attrs_str} />"
    else:
        html = f"<{tag}{attrs_str}>"
    return html, attrs_dict



from setuptools import setup, find_packages

with open("README.md", encoding="utf8") as f:
    long_description = f.read()

setup(
    name="gazpacho",
    version="1.0",
    description="The simple, fast, and modern web scraping library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords=["web scraping", "BeautifulSoup", "requests"],
    url="https://github.com/maxhumber/gazpacho",
    author="Max Humber",
    author_email="max.humber@gmail.com",
    license="MIT",
    packages=find_packages(),
    extras_require={
        "dev": ["black>=20.8b1", "mypy>=0.782", "portray>=1.4.0", "pytest>=6.0.2"],
    },
    python_requires=">=3.6",
    setup_requires=["setuptools>=38.6.0"],
)



import json
from typing import Dict, Optional, Union
from urllib.error import HTTPError as UrllibHTTPError
from urllib.parse import quote, urlencode, urlsplit, urlunsplit
from urllib.request import build_opener


class HTTPError(Exception):
    def __init__(self, code: int, msg: str) -> None:
        self.code = code
        self.msg = msg

    def __str__(self):
        return f"{self.code} - {self.msg}"


def sanitize(url: str) -> str:
    """\
    Sanitize and format a URL

    Arguments:

    - url: target page
    """
    scheme, netloc, path, query, fragment = urlsplit(url)
    if not scheme:
        scheme, netloc, path, query, fragment = urlsplit(f"http://{url}")
    path = quote(path)
    url = urlunsplit((scheme, netloc, path, query, fragment))
    return url


def get(
    url: str,
    params: Optional[Dict[str, str]] = None,
    headers: Optional[Dict[str, str]] = None,
) -> Union[str, dict]:
    """Retrive url contents

    Params:

    - url: target page
    - params: GET request payload
    - headers: GET request headers

    Example:

    ```
    get('https://httpbin.org/anything', {'soup': 'gazpacho'})
    ```
    """
    url = sanitize(url)
    opener = build_opener()
    if params:
        url += "?" + urlencode(params)
    if headers:
        for h in headers.items():
            opener.addheaders = [h]
    if (headers and not headers.get("User-Agent")) or not headers:
        UA = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.16; rv:80.0) Gecko/20100101 Firefox/80.0"
        opener.addheaders = [("User-Agent", UA)]
    try:
        with opener.open(url) as response:
            content = response.read().decode("utf-8")
            if response.headers.get_content_type() == "application/json":
                content = json.loads(content)
    except UrllibHTTPError as e:
        raise HTTPError(e.code, e.msg) from None
    return content
